# Copyright (C) 2020 Ludvig Larsson aka Valmond ludviglarsson@protonmail.com
# This code is licensed under the terms of the GNU Public License, version 3 or later (GPL-3.0-or-later)

import sys
import os

#from Cryptodome.Cipher import AES
#from Cryptodome.Util import Counter
#from Cryptodome import Random
import binascii
import os.path
import tarfile
from tarfile import GNU_FORMAT  # To remove those pesky @PaxHeader infos in our tar.gz
import re
from contextlib import contextmanager

import database_connection as dc
import database_query as dq

from encryption import CipherHelper
import locator


# This loads and encrypts data, inserts it into the table 'data_owned'
# stores encrypted data in folder '/data/local/' as it is our owned data.
# Later on, the scheduler will ask other nodes for swicharoo (e.g storing our data at their place, and their data at our place in /data/$nodenr$/)

def data_exists(data_name):
    connection = dc.get_connection()
    cursor = connection.cursor()
    cursor.execute("SELECT key FROM data_owned WHERE NAME = '" + data_name + "' LIMIT 1")  # todo: secure this
    rows = cursor.fetchall()
    # Does a database entry exists?
    return len(rows) > 0


def chomp(x):
    if x.endswith("\r\n"):
        return x[:-2]
    if x.endswith("\n") or x.endswith("\r"):
        return x[:-1]
    return x


# Inserts new data or updates existing data, does nothing if the data is (exactly) the same.
# rsa_pub_key, address_translation_file are only used when creating new data
# is_tar_gz_folder is needed for when we compare tar.gz files
# Note: If you want to insert data, do not use this function directly (or be sure to know what you are doing), use insert_file and insert_folder instead.
def insert_our_data(name, filefolder, file, rsa_pub_key=None, address_translation_file=0, is_tar_gz_folder=False):
    # Insert in database
    print('Insert new/update data:')
    print('Name:', name)
    print('Folder:', filefolder)
    print('File:', file)

    filefolder = filefolder.replace('\\', '/')
    file = file.replace('\\', '/')

    connection = dc.get_connection()

    #dir_path = os.path.dirname(os.path.realpath(__file__))
    filepath = os.path.join(filefolder, file)

    if not os.path.exists(filepath):
        print("#Error: cannot find file", filepath)
        return False

    data = None
    with open(filepath, 'rb') as f:
        data = f.read()

    if data is None:
        print('File not found')
        return False

    # Does a database entry exists?
    update_data = data_exists(name)

    # Make a 'local' folder if it doesn't already exist
    directory = os.path.join(locator.module_path(), 'data/local/')
    if not os.path.exists(directory):
        os.makedirs(directory)

    # Make the encrypted result files name
    aes_file_name = directory + name + '.aes'

    # The nonce is a number usable only once for AES encryption
    # It can freely be distributed and does not be hidden like the key.
    # Its uniqueness is only needed per each specific AES key in CFB or CTR mode,
    # but it needs to be unique in every block (encrypted with a specific key),
    # not only once for each starting block for example.
    # We use AES256-CTR, and it can thus only encrypt a total of around 64Gib/68GB per key.
    # When that amount of data has been transferred, it is no longer safe to use.
    # For example, if your data is 1MB you can only change it around 65000 times.

    next_safe_nonce = 0

    if update_data:
        cursor = connection.cursor()
        cursor.execute("SELECT key, iv, next_safe_iv FROM data_owned WHERE NAME = ?", (name, ))
        rows = cursor.fetchall()
        # Does a database entry exists?
        if len(rows) <= 0:
            print("#Error database problems...")
            return False
        # Get IV, KEY
        row = rows[0]
        key = bytes.fromhex(chomp(row[0]))
        old_used_nonce = int(row[1])
        next_safe_nonce = int(row[2])

        # Make the cipher helper that will encrypt our data, starting at an unused nonce:
        ch = CipherHelper(aes_key=key, nonce=next_safe_nonce)

        # Now check if the data is new, otherwise just bail out
        new_data_size = len(data)
        aes_file_size = os.path.getsize(aes_file_name)

        # Note: There are 8 random numbers at the beginning in AES files to anonymize data
        if new_data_size + 8 == aes_file_size:
            with open(aes_file_name, 'rb') as f:
                aes_file_data = f.read()
                if file is not None:
                    # Make a ciperhelper that will decrypt the old stored data:
                    # Note: we cannot just compare the .aes files, as there is an anonymisation done to them. We must decrypt.
                    ch_tmp = CipherHelper(aes_key=key, nonce=old_used_nonce)
                    old_data = ch_tmp.decrypt(aes_file_data)

                    same = False

                    if is_tar_gz_folder:
                        # Folders are stored and compressed in the tar gz format.
                        # To compare two tar.gz files, we need to skip the timestamp, which
                        # lives in byte 4 to 7
                        # Also, of course, we still need to take into account the 8 random numbers in the data stored on disk
                        same_start = old_data[8:12] == data[:4]  # The 4 first bytes(magic number, but also compression method and file flags so we need to check them)
                        same_end = old_data[8 + 8:] == data[8:]  # Skip the 4 bytes for the timestamp
                        same = same_start and same_end
                    else:
                        # Again those pesky 8 random numbers must be taken into account, but that is all, just compare all the rest
                        same = old_data[8:] == data

                    if same:
                        print('Skip updating [' + name + '] as the data is exactly the same as the existing version')
                        return True

    else:
        # Completely new data, make a key with a nonce starting at 0
        ch = CipherHelper()

    nonce_used = ch.get_nonce()

    # Encrypt with AES-CTR
    encrypted_data = ch.encrypt(data, True)

    # Update the nonce
    next_safe_nonce = ch.get_nonce()


    # Store the encrypted data in the data folder so listener can serve it to clients:

    if update_data:
        print('Update the data')
    else:
        print('Write the new data')

    with open(aes_file_name, 'wb') as f:
        f.write(encrypted_data)

    AEShex = ch.aes_key.hex()
    ##IVhex = ch.raw_iv.hex()

    cursor = connection.cursor()
    if update_data:
        cursor.execute("UPDATE data_owned SET data_version=data_version+1, originalfilename = ?, iv = ?, next_safe_iv = ? WHERE name = ? ;",
                       (file, str(nonce_used), str(next_safe_nonce), name))
        # Check if there is a discrepancy instead:
        # if rsa_pub_key is not None or address_translation_file != 0:
        #    print("#Warning: Do not update data using rsa_pub_key or address_translation_file, they are skipped")
    else:
        cursor.execute("INSERT INTO data_owned(NAME, address_translation_file, ORIGINALFILENAME, key, iv, next_safe_iv, PUBKEY)"
                       "VALUES(?, ?, ?, ?, ?, ?, ?);",
                       (name, address_translation_file, file, AEShex, str(nonce_used), str(next_safe_nonce), rsa_pub_key))
    connection.commit()
    return True


# Accessible as python function:
def insert_file(data_name, folder, file_name, address_translation_file=0):
    _, pub = dq.get_rsa_hex_keys()
    insert_our_data(data_name, folder, file_name, pub, address_translation_file)


# Easy peasy change dir and back
@contextmanager
def cwd(path):
    oldpwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(oldpwd)


# Accessible as python function:
def insert_folder(data_name, base_folder, folder_to_add, regex, silent=False):
    folder = folder_to_add # os.path.join(base_folder, folder_to_add)
    if not silent:
        print('Insert folder:', folder)

    # Go to the base folder so that the ter file won't have them in the filenames
    with cwd(base_folder):
        # get all the files in 'folder' and zip it:
        full_data_name = data_name + '.tar.gz'

        with tarfile.open(full_data_name, "w:gz", format=GNU_FORMAT) as tar_handle:
            for root, _, files in os.walk(folder):
                for file in files:
                    file_path = os.path.join(root, file)
                    if re.search(regex, file_path):
                        if not silent:
                            print('# ADD:', file_path)
                        tar_handle.add(file_path)  #, arcname=os.path.basename(os.path.join(root)))
                    else:
                        if not silent:
                            print('# SKIP:', file_path)

    _, pub = dq.get_rsa_hex_keys()
    # name, filefolder, file, rsa_pub_key
    insert_our_data(data_name, base_folder, data_name + '.tar.gz', pub, is_tar_gz_folder=True)
    # delete tar.gz file
    os.remove(os.path.join(base_folder, full_data_name).replace('\\', '/'))

